import tomlkit
from contextlib import contextmanager
from os import PathLike


@contextmanager
def edit_toml(file_path: str):
    with open(file_path) as f:
        loaded = tomlkit.load(f)

    try:
        yield loaded
    finally:
        with open(file_path, 'w') as f2:
            tomlkit.dump(loaded, f2)


def load_toml(file_path: str | PathLike):
    with open(file_path) as f:
        return tomlkit.load(f)


def save_toml(obj: dict, file_path: str | PathLike):
    with open(file_path, 'w') as f:
        tomlkit.dump(obj, f)



from .base import CoreDatabase
from cryptography.fernet import Fernet
from typing import Optional, Union
from os import environ as env


class AutoEncodeFernet(Fernet):
    def decrypt(self,
                token: Union[bytes, str, memoryview],
                ttl: Optional[int] = None
                ) -> str:
        """Modified decrypt method for processing memory-view object from database to a decoded str"""
        token = bytes(token) if isinstance(token, memoryview) else token
        return super().decrypt(token, ttl).decode()

    def encrypt(self, data: Union[str, bytes]) -> bytes:
        """Modified encryption method for accepting str objects"""
        data = data.encode() if isinstance(data, str) else data
        return super().encrypt(data)


class CryptoDatabase(CoreDatabase):
    @classmethod
    def env_key(cls, cxn_config: dict):
        """Method that allows for database cryptography key to be automatically
        pulled from environmental variable "CRYPTO_KEY" """
        crypto_key = env.get('CRYPTO_KEY')
        if not crypto_key:
            raise Exception('Environmental variable "CRYPTO_KEY" not found.')
        return cls(cxn_config, crypto_key)

    def __init__(self, cxn_config: dict, crypto_key: str | bytes):
        """
        Elevated database interface that allows for automatic encrypting and decrypting of
        specified database entries.

        Parameters
        ----------
        cxn_config : dict
            dictionary containing all database connection parameters (ie host, port, password, database)
        crypto_key : str | bytes
            32 character url-safe base64 encoded key to be used with encryption and decryption

        """
        super().__init__(cxn_config)
        self._crypt = AutoEncodeFernet(crypto_key)

    # TODO: instead parse tables and find which are marked as encrypted OR ask for all keys
    def crypt(self, p: dict | list[dict], client_key='client', secret_key='secret', rf_tk_key='refresh_token'):
        # save type for return value
        is_dict = isinstance(p, dict)
        # homogenize input
        p = [p] if is_dict else p

        # decrypt or encrypt method set on first use
        func = None
        # reduce to necessary keys
        ks = [k for k in [client_key, secret_key, rf_tk_key] if k in p[0]]
        for x in p:
            for k in ks:
                if v := x.get(k):
                    # set method if needed
                    if not func:
                        # auto-detects method based on type, encrypted data will be bytes
                        func_name = 'decrypt' if isinstance(v, memoryview) else 'encrypt'
                        func = getattr(self._crypt, func_name)

                    # gets method and calls with value
                    x[k] = func(v)
        return p[0] if is_dict else p



from fastapi import Request
from fastapi.responses import HTMLResponse, Response, RedirectResponse
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.types import ASGIApp, Receive, Scope, Send
import typing

from .client import AuthClient
from .etc import RevokedAuth


# TODO: ADD base middleware super, implement for redirect jank
class AuthMiddleware:
    def __init__(
            self,
            app: ASGIApp,
            client: AuthClient,
            available_scopes: typing.Optional[typing.Sequence[str]] = None,
            restricted_paths: typing.Optional[typing.Sequence[str]] = None,

    ) -> None:
        self.app = app
        self.client: AuthClient = client
        self.available_scopes = available_scopes
        self.restricted_paths = restricted_paths
        self.mingle = None  # Check for redis and other clients

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        # skip non-http requests like websockets
        if scope['type'] != 'http':
            await self.app(scope, receive, send)

        else:
            path: str = scope.get('path')
            request = Request(scope=scope)

            # catch authentication sdk endpoints
            if path.startswith('/bsAuth/callback'):
                response: Response
                ref = request.cookies.get('bt_rf')
                pack = self.client.full_fetch(ref)

                self.client.sessions[ref] = {'pending': False, 'revalidated': False, **pack['azp']}
                # language=html
                callback_html = """
                <!DOCTYPE html><html lang="en">
                <head><meta charset="UTF-8"><title>Authorized</title></head>
                <body>Good auth.<script>window.close()</script></body>
                </html>
                """

                response = HTMLResponse(callback_html, status_code=200)

                if self.client.sessions[ref]['user_id']:
                    response.set_cookie(key='user_id', value=self.client.sessions[ref]['user_id'], samesite='none',
                                        secure=True, expires=2147483647)
                else:
                    response.delete_cookie('bt_rf')

                await response(scope, receive, send)

            elif path.startswith('/bsAuth/login'):
                pending = self.client.build_auth()
                self.client.sessions[pending['ref']] = {'pending': True, 'user_id': None}
                resp = Response(content=pending['url'], status_code=200)
                resp.set_cookie('bt_rf', pending['ref'], httponly=True, secure=True, samesite='none',
                                expires=2147483647)
                await resp(scope, receive, send)

            elif path.startswith('/bsAuth/logout'):
                print(request)

                # language=html
                page_html = """
                    <!DOCTYPE html><html lang='en-us'>
                    <head><title>Logged Out</title></head>
                    <body>
                        <div><p>
                            Logout successful. 
                            Redirecting in <span id='countdown'>3</span> seconds. 
                            Click <a href='/'>here</a> to jump immediately.
                        </p></div>
                        
                        <script>
                            const elem = document.getElementById('countdown');
                            const updateOrJump = () => {
                                let curTime = Number(elem.innerText);
                                if (curTime > 0) {
                                    curTime--;
                                    elem.innerText = String(curTime);
                                    setTimeout(updateOrJump, 1000);
                                } else {
                                    window.location = '/';
                                }
                            }
                            window.onload = () => updateOrJump();
                        </script>
                    </body>
                    </html>
                    """

                # script-based redirection response for fallback logout,
                # typically this endpoint will be called via fetch, in which case the logout
                # flow will execute and the above redirect script can be ignored and the
                # response status can instead be used to trigger a callback function.
                response = HTMLResponse(content=page_html, status_code=200)

                # auth server token invalidation when active cookie present
                if ref := request.cookies.get('bt_rf'):
                    self.client.invalidate_ref(ref)
                    # cookie modification attached to the response
                    response.delete_cookie('bt_rf')

                # handle temp and complete failed logouts by clearing user
                if 'user_id' in request.cookies:
                    response.delete_cookie('user_id')

                await response(scope, receive, send)

            else:
                # Remove revoked tokens if present
                try:
                    self.client.user(request)
                    await self.app(scope, receive, send)
                except RevokedAuth:
                    response = RedirectResponse(path)
                    response.delete_cookie('bt_rf')
                    response.delete_cookie('user_id')
                    await response(scope, receive, send)



from time import time
from datetime import datetime, UTC, timedelta
# from warnings import deprecated


def rni():
    """ms since epoch. js equiv"""
    return int(time() * 1000)


def rn(tz=UTC):
    """Aware datetime defaulting to UTC"""
    return datetime.now(tz)


def age(dt: datetime):
    return datetime.now(dt.tzinfo) - dt


def stale(dt: datetime | None, max_age: timedelta):
    if not dt:
        return True
    return age(dt) > max_age



from typing import MutableMapping, Union, Optional, Any, TypedDict
import json
from redis import Redis
from os import environ as env


class BaseConfig(TypedDict):
    host: str
    port: int
    password: Optional[Any]
    db: Optional[int]


class RedisMapping(MutableMapping):
    @classmethod
    def from_env(cls, db_idx: int):
        ks = ['RM_HOST', 'RM_PORT', 'RM_PASS']
        not_present = [x for x in ks if x not in env]
        if not_present:
            print(f'Redis Mapping from env using default values for {not_present}')

        return cls({
            'host': env.get('RM_HOST', 'localhost'),
            'port': env.get('RM_PORT', 6379),
            'password': env.get('RM_PASS', None)
        }, db_idx)

    def __init__(self, base_config: Union[BaseConfig, dict], db_idx: int):
        self.config = base_config | {'db': db_idx}
        self.cxn = Redis(**self.config)

    def __getitem__(self, item):
        if found := self.cxn.get(item):
            try:
                return json.loads(found)
            except UnicodeDecodeError:
                return found
        raise KeyError

    def __len__(self):
        return self.cxn.info().get('db0', {'keys': 0})['keys']

    def __setitem__(self, key, value):
        if not (isinstance(value, str) or isinstance(value, bytes)):
            value = json.dumps(value)
        self.cxn.set(key, value)

    def __delitem__(self, key):
        if self.cxn.exists(key):
            self.cxn.delete(key)

    def __iter__(self):
        for x in self.cxn.keys():
            yield x





def fracture(lst: list, size: int | None = 100, with_idx=False) -> list | tuple[list, int]:
    for i in range(0, len(lst), size):
        if with_idx:
            yield lst[i:i + size], i
        else:
            yield lst[i:i + size]



from typing import TypedDict, NotRequired

# enabling type hints for misc stuff


class ConnectionParams(TypedDict):
    """
    - *dbname*: the database name
    - *database*: the database name (only as keyword argument)
    - *user*: user name used to authenticate
    - *password*: password used to authenticate
    - *host*: database host address (defaults to UNIX socket if not provided)
    - *port*: connection port number (defaults to 5432 if not provided)
    """
    pass



from contextlib import contextmanager
from typing import Generator, Optional, Any
from psycopg_pool import ConnectionPool
from psycopg import connect
from psycopg import cursor
from psycopg.errors import DeadlockDetected
from os import environ as env


def get_comparator(v) -> str | tuple:
    if isinstance(v, str):
        if 'null' in v:
            return ('is not', None) if 'not' in v else ('is', None)
        return 'like'
    elif isinstance(v, int):
        return '='

    return 'is'


class CoreDatabase:
    """
    Base interface for postgres database, includes methods for
    querying, selection, insertion, and deletion
    """

    @staticmethod
    def cxn_params_from_env():
        return {
            'host': env.get('DB_HOST', 'localhost'),
            'port': env.get('DB_PORT', 5432),
            'dbname': env.get('DB_DATABASE', 'postgres'),
            'user': env.get('DB_USER', 'postgres'),
            'password': env.get('DB_PASSWORD', None)
        }

    @classmethod
    def from_env(cls,
                 etc: Optional[dict] = None,
                 verbose: Optional[bool] = False,
                 **kwargs
                 ):
        """
        Automatically pull basic connection parameters [host, port, database, user, and password]
        from the environmental variables (caps and prefixed with DB_ ex: DB_HOST) and return a database instance.
        Can be provided with a dict of additional connection parameters which will override
        values from the environment or be appended if not present.

        Parameters
        ----------
        etc : dict
            additional connection parameters used to override/add to environmental variables.
        verbose : bool
            provides variable parsing output
        Returns
        -------
        CoreDatabase
            A core database instance with connection parameters from the environment.
        """

        def v_print(*s):
            if verbose:
                print(*s)

        v_print('Building database connection parameters from env...')
        missing = [x for x in ['DB_HOST', 'DB_PORT', 'DB_DATABASE', 'DB_USER', 'DB_PASSWORD'] if x not in env]
        v_print('env missing: ', missing, 'using default values for these fields.')

        # build base connection parameters, filling where needed
        cxn = {
            'host': env.get('DB_HOST', 'localhost'),
            'port': env.get('DB_PORT', 5432),
            'dbname': env.get('DB_DATABASE', 'postgres'),
            'user': env.get('DB_USER', 'postgres'),
            'password': env.get('DB_PASSWORD', None)
        }

        # merge with overrides if needed
        if etc:
            v_print(f'received additional config. '
                    f'adding: {[x for x in etc if x not in cxn]} '
                    f'overwriting: {[x for x in etc if x in cxn]}')

            cxn = cxn | etc

        v_print(f'Finalized connection parameters: ', cxn)

        return cls(cxn, **kwargs)

    def __init__(self, cxn_config: dict, pool_config: Optional[dict] = None, **kwargs):
        """

        Parameters
        ----------
        cxn_config : dict
            connection parameters fed to Psycopg connection init (ie host, port, user, password, dbname, etc)
            See https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-PARAMKEYWORDS

        kwargs : dict
            keyword argument added for super classing support for class methods
        """
        # parse connection parameters from ini file and merge
        # with overrides if provided
        self.db = None
        self._tables = None
        self.config = cxn_config
        self.params = {'min_size': 2, 'open': True} | (pool_config if pool_config else {})
        self.pool = ConnectionPool(kwargs=self.config, **self.params)
        self.__post_init__()

    def __post_init__(self):
        self.db = self.query('SELECT current_database()', as_tuple=True)[0]

    def query(self,
              query: str,
              vals: Any = None,
              as_tuple: bool | None = False,
              no_resp: bool | None = False,
              tidx: int | None = None,
              deadlock_retries: int | None = 1,
              **kwargs
              ) -> list:
        """
        Committed database query execution

        Parameters
        ----------
        query : str
            the command to be executed

        vals : tuple
            parameterized values to be included with query if applicable
            (default=None)

        as_tuple : bool
            modify return value to be a tuple list instead of dict

        no_resp : bool
            whether to attempt to fetch and parse execution output,
            ie from a select query (default=False)

        tidx : int
            allows tuple return to be preparsed to the given index
            [commonly] tidx=0 will return values instead of a tuple singleton

        deadlock_retries : int
            times to retry query execution upon deadlock (Default = 1)

        Returns
        -------
        list
            Record-like dict list when resp specified
        """
        # auto pack to tuple if needed
        if vals and not (isinstance(vals, dict) or isinstance(vals, tuple)):
            vals = vals,

        fetched, cols = [], []
        if tidx is not None:
            as_tuple = True

        # execute with temp cursor
        with self.pool.connection() as cxn:
            with cxn.cursor() as csr:
                csr: cursor.Cursor
                try:
                    csr.execute(query, vals)
                except DeadlockDetected as e:
                    if deadlock_retries:
                        return self.query(
                            query=query, vals=vals, as_tuple=as_tuple, tidx=tidx,
                            deadlock_retries=deadlock_retries-1, **kwargs)
                    else:
                        raise e
                try:
                    if not no_resp:
                        # grab columns for record formation
                        if not as_tuple:
                            cols = tuple(x[0] for x in csr.description)
                        # retrieve values
                        fetched = csr.fetchall()
                except TypeError:
                    cols = None
                    fetched = None

        # cursor returned to pool before parsing, if required.
        if fetched:
            if cols:
                return [{k: v for k, v in zip(cols, row)} for row in fetched]

            elif as_tuple:
                if tidx is not None:
                    return [x[tidx] for x in fetched] if fetched else []

                return [x for x in fetched]

        elif not no_resp:
            return kwargs.get('default', [])

    @property
    def tables(self):
        if not self._tables:
            q = "select table_name from information_schema.tables where table_schema = 'public'"
            resp = self.query(q, as_tuple=True)
            self._tables = [x[0] for x in resp if x] if resp else []
        return self._tables

    @property
    def table_sizes(self):
        # language=sql
        q = """
        SELECT
            table_name,
            pg_size_pretty(table_size) AS table_size,
            pg_size_pretty(indexes_size) AS indexes_size,
            pg_size_pretty(total_size) AS total_size
        FROM (
            SELECT
                table_name,
                pg_table_size(table_name) AS table_size,
                pg_indexes_size(table_name) AS indexes_size,
                pg_total_relation_size(table_name) AS total_size
            FROM (
                SELECT ('"' || table_schema || '"."' || table_name || '"') AS table_name
                FROM information_schema.tables
            ) AS all_tables
            ORDER BY total_size DESC
        ) AS pretty_sizes;
        """
        return self.query(q)

    def select(self, table: str, field='*', where: dict | None = None, limit=None) -> list | None:
        """
        Fetch rows from a database table within specified fields

        Parameters
        ----------
        table : str
            a valid table reference within the initialized database
        field : str
            valid sql field selector str
        where : dict
            passable dict of where results are query results of matching k, v pairs
        limit : int
            set max items, None returns all (default None)

        Returns
        -------
        list:
            Record-like list of dicts keyed to table fields

        """
        if table not in self.tables:
            raise KeyError(f'Table "{table}" is unavailable in current database.')

        # set where params
        passed_vars = []
        if not where:
            where = ''
        else:
            conditions = []
            for k, v in where.items():
                comparator = get_comparator(v)
                if isinstance(comparator, tuple):
                    comparator, v = comparator

                conditions.append(f'{k} {comparator} %s')
                passed_vars.append(v)
            where = f" where {' and '.join(conditions)}"

        # prepare parameterized values to be passed to query when present
        passed_vars = None if not passed_vars else tuple(passed_vars)

        # modify or remove limit if specified
        limit = f' limit {limit}' if limit else ''

        # form full query str
        q = f'select {field} from {table}{where}{limit};'

        # identify if values should be returned as a dict or singleton list
        single = field != '*' and field.find(',') == -1

        # execute, requesting a tuple (field parsing skipped) if allowable
        resp = self.query(q, vals=passed_vars, as_tuple=single)

        # early return to escape NoneType iteration error
        if not resp:
            return []

        # forcing return as a list, for some reason (probably the cursor being
        # a generator), its returning and iterable.
        return [x[0] for x in resp] if single else [x for x in resp]

    def insert(self, table: str, data: dict | list, upsert_on: str | None = None, fill_na=False) -> None:
        """
        Fast parameterized insertion

        Parameters
        ----------
        table : str
            target for data entry
        data : dict
            insertion data in record format, with keys matching
            specified table fields.
        upsert_on : str
            unique column key for which collisions result in an update (default = None)
        fill_na : bool
            attempts to fill missing keys with None (default = False)
        Raises
        ------
        psycopg2.errors.UndefinedTable
            on invalid table
        psycopg2.errors.UndefinedColumn
            on invalid fields

        Returns
        -------
        None
            None
        """

        if not data:
            raise ValueError

        # force list
        data = [data] if isinstance(data, dict) else data

        # data uniformity is NOT assumed, finds insertion keys
        kls = tuple(max(data, key=len).keys())

        # kls = tuple(data[0].keys())

        # sql formatted table keys generated
        ks = f'({", ".join(kls)})'

        # build correct length parameterized variable (%s, %s, ...) strings
        kvs = f'({", ".join("%s" for _ in kls)})'

        # set entry variables
        vks = ', '.join(kvs for _ in data)

        # grab all values and flatten to tuple
        if fill_na:
            vs = tuple(p.get(k) for p in data for k in kls)
        else:
            vs = tuple(p[k] for p in data for k in kls)
        # if previous data non-uniformity handling is implemented,
        # missing keys must instead be filled with:
        # vs = tuple(z for x in data for z in tuple(data[x].get(k, None) for k in kls))

        # set collision result
        ex_ks = f"({', '.join(f'EXCLUDED.{k}' for k in kls)})"
        collision = 'DO NOTHING' if not upsert_on else f"({upsert_on}) DO UPDATE SET {ks} = {ex_ks}"

        # full query formation
        q = f"insert into {table} {ks} VALUES {vks} ON CONFLICT {collision};"

        # submit for execution with parameterized values
        self.query(q, vs, no_resp=True)

    def delete(self, table: str, where: dict):
        """
        Removes item(s) from the specified table.

        Parameters
        ----------
        table : str
            table to execute deletion on
        where : dict[str, str] | dict[str, tuple[str, str]]
            dictionary of key value pairs to match for deletion, comparator is
            automatically chosen based on data type, but can also be forced by
            replacing the value with a tuple containing the comparator and value.

        Examples
        --------
        `standard call` : delete(table='foo', where={'bar': 6})
             ==sql=> delete from foo where bar = 6;
        `multiple where` : delete(table='foo', where={'bar': 8, 'buzz': 'yellow'})
            ==sql=> delete from foo where bar = 8 and buzz like yellow;
        `forcing comparator` : delete(table='foo', where={'bar': 8, 'buzz': ('=', 'yellow')})
         ==sql=> delete from foo where bar = 8 and buzz = yellow;

        """
        conditions = []
        passed_vars = []
        for k, v in where.items():
            if isinstance(v, tuple):
                if len(v) != 2:
                    raise Exception('comparator forcing must be formatted as (comparator, value)')
                else:
                    comparator, v = v
            else:
                comparator = get_comparator(v)
            conditions.append(f'{k} {comparator} %s')
            passed_vars.append(v)
        where_conditions = f"{' and '.join(conditions)}"
        q = f'delete from {table} where {where_conditions}'
        self.query(q, vals=tuple(passed_vars), no_resp=True)



from typing import Literal, TypedDict


AvailableEndpoints = Literal[
    'activeKey',
    'credential',
    'retrieveToken',
    'logout'
]


class FailedAuth(Exception):
    """Authentication fetch failed"""


class RevokedAuth(Exception):
    """Access to provided token was revoked"""


class HostileEnvironment(Exception):
    """Missing environmental keys"""


class PreppedAuth(TypedDict):
    """Object containing the authorization url for client login and its associated reference string."""
    ref: str
    url: str



from typing import Optional
import pickle
import requests
import os
from os import PathLike
import json
import copy
from datetime import timedelta, datetime


class CachedRequester:
    def __init__(
            self,
            path: str | PathLike = '.birdwell/requests/cache.json',
            stale_timeout: timedelta | None = None,
            verbose=False
    ):
        if not str(path).endswith('.json'):
            raise ValueError('invalid cache file loc')
        self.path = path
        self._cache = None
        self.desync = True
        self.stale_timeout = stale_timeout
        self.verbose = verbose
        # atexit.register(self.dump_cache)

    @property
    def cache(self):
        if self._cache is None:
            if os.path.exists(self.path):
                with open(self.path) as f:
                    self._cache = json.load(f)
            else:
                self._cache = {}
        return self._cache

    @cache.setter
    def cache(self, val):
        self._cache = val
        self.dump_cache()

    def dump_cache(self):
        with open(self.path, mode='w') as f:
            json.dump(self._cache, f, indent=2)

    def is_stale(self, endpoint):
        if self.stale_timeout is None:
            return False

        return (datetime.now() - self.stale_timeout).timestamp() > self.cache[endpoint]['fetch_ts']

    def get_json(self, endpoint):
        if endpoint not in self.cache or self.is_stale(endpoint):
            if self.verbose:
                print(f'Requesting Resource: "{endpoint}"')
            self.cache |= {endpoint: {
                'data': requests.get(endpoint).json(),
                'fetch_ts': datetime.now().timestamp()
            }}

        return copy.deepcopy(self.cache[endpoint]['data'])



import json
import pickle as pk
from typing import Any, Generator

from .base import CoreDatabase
from collections.abc import MutableMapping


class MappedDatabase(CoreDatabase, MutableMapping):
    """
    Interface that allows for a database table to be interacted with as if it were a dictionary object.
    """
    # TODO: add a mode selector pickle v json
    def __init__(self, cxn_config: dict, table: str, pickle=False):
        self.bound_table = table
        self._pickle = pickle
        if self._pickle:
            print('WARNING: Pickle mode is active. Not for use in production. Learn more about pickle '
                  'vulnerabilities here: https://docs.python.org/3/library/pickle.html')
        self._vk = 'val_p' if self._pickle else 'val'
        super().__init__(cxn_config)

    def __repr__(self):
        return f'Mutable mapping interface bound to table "{self.bound_table}" of database "{self.db}".'

    def __post_init__(self):
        super().__post_init__()
        print(f'ensuring {self.bound_table} in database:', self.tables)
        if self.bound_table not in self.tables:
            print('generating table')
            tq = f"""
            create table {self.bound_table} (
                var varchar(255) primary key,
                val jsonb,
                val_p bytea,
                ts timestamp default current_timestamp
            );
            """

            # add table
            self.query(tq, no_resp=True)

            # clear cached table names
            self._tables = None

    def dumps(self, v) -> bytes | str:
        if self._pickle:
            return pk.dumps(v)
        return json.dumps(v)

    @staticmethod
    def loads(v) -> Any:
        if isinstance(v, memoryview):
            return pk.loads(v)
        return v

    def __setitem__(self, __k: Any, __v: Any) -> None:
        self.insert(self.bound_table, {'var': __k, self._vk: self.dumps(__v)}, upsert_on='var')

    def __delitem__(self, __v: Any) -> None:
        self.delete(self.bound_table, {self._vk: self.dumps(__v)})

    def __getitem__(self, __k: Any) -> Any:
        resp = self.select(self.bound_table, self._vk, where={'var': __k}, limit=1)

        # throw missing
        if not resp:
            raise KeyError(f'{__k} not present.')

        # enable de-pickling via parse
        return self.loads(resp[0])

    def __len__(self) -> int:
        resp = self.select(self.bound_table, f'count({self._vk})')
        return resp[0] if resp else 0

    def __iter__(self) -> Generator:
        for x in self.select(self.bound_table, 'var', where={self._vk: 'not null'}):
            yield x



import random
import string
from typing import Optional, Literal


def f_up(txt: str):
    """Capitalize first letter of string"""
    return txt[0].upper() + txt[1:]


Case = Literal['upper', 'lower', 'both']


def ascii_code(length: Optional[int] = 6,
               case: Optional[Case] = 'both',
               digits: Optional[bool] = True
               ):
    pools = {
        'both': string.ascii_letters,
        'lower': string.ascii_lowercase,
        'upper': string.ascii_uppercase
    }
    return ''.join(random.choices(pools[case] + string.digits if digits else [], k=length))


class TextColor:
    PURPLE = '\033[95m'
    BLUE = '\033[94m'
    CYAN = '\033[96m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    END = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'




import json
import base64
import requests

from typing import Optional
from os import environ as env

from fastapi import Request, Depends

from cryptography.hazmat.primitives.asymmetric.padding import PSS, MGF1
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.hazmat.primitives import hashes, serialization as cereal
from cryptography.exceptions import InvalidSignature

from .etc import AvailableEndpoints, FailedAuth, PreppedAuth, RevokedAuth, HostileEnvironment


class AuthClient:
    @classmethod
    def from_env(cls,
                 *void,
                 client_id: Optional[str] = None,
                 client_key: Optional[str] = None,
                 redirect_uri: Optional[str] = None,
                 allow_revalidation: Optional[bool] = False
                 ):
        """

        Parameters
        ----------
        void : any
            positional arguments not allowed.
        client_id : str
            optional client id override
        client_key : str
            optional api key override
        redirect_uri : str
            optional redirect overrides
        allow_revalidation : bool
            allow known keys to be revalidated instead of rebuilt

        Returns
        -------

        """

        client = client_id if client_id else env.get('AUTH_CLIENT_ID')
        key = client_key if client_key else env.get('AUTH_API_KEY')
        uri = redirect_uri if redirect_uri else env.get('AUTH_REDIRECT_URI')

        missing_envs = []
        if not client:
            missing_envs.append('AUTH_CLIENT_ID')
        if not key:
            missing_envs.append('AUTH_API_KEY')
        if not uri:
            missing_envs.append('AUTH_REDIRECT_URI')

        if missing_envs:
            raise HostileEnvironment(f'Cannot build auth client from environment. Missing variables: {missing_envs}.')

        return cls(client_id=client, client_key=key, redirect_uri=uri, allow_revalidation=allow_revalidation)

    def __init__(self, client_id: str, client_key: str, redirect_uri: str, allow_revalidation=False):
        self.id = client_id
        self.key = client_key
        self.redirect_uri = redirect_uri
        self.allow_revalidation = allow_revalidation
        self.auth = {'client_id': self.id, 'api_key': self.key}
        self.server = 'https://auth.brd.cx'
        self.core_encoding = ''
        self._public_key: RSAPublicKey | None = None
        self.padding = PSS(mgf=MGF1(hashes.SHA3_224()), salt_length=15)
        self.sha256 = hashes.SHA256()
        self.sessions = {}

    def user(self, request: Request) -> dict:
        bt_rf = request.cookies.get('bt_rf')
        current_user = request.cookies.get('user_id')
        if not bt_rf:
            user = None
        else:
            if not (user := self.sessions.get(bt_rf)) and self.allow_revalidation and current_user:
                pre_authed = self.full_fetch(bt_rf)
                if not pre_authed:
                    # invalidate?
                    user = None
                elif current_user == pre_authed['azp']['user_id']:
                    user = self.sessions[bt_rf] = {'pending': False, 'revalidated': True, **pre_authed['azp']}
        return user

    @staticmethod
    def b64_encode(data: str | bytes) -> str:
        return base64.urlsafe_b64encode(data.encode() if isinstance(data, str) else data).decode()

    @staticmethod
    def b64_decode(code: str) -> bytes:
        return base64.urlsafe_b64decode(code.encode())

    def _request(self,
                 endpoint: AvailableEndpoints,
                 params: dict | None = None,
                 body: dict | None = None,
                 raw=False
                 ):
        url = f'{self.server}/{endpoint}'
        if body:
            resp = requests.post(
                url=url,
                headers={'Content-Type': 'application/json'},
                json=body)
        else:
            resp = requests.get(url=url, params=params)
        if raw:
            return resp

        if resp.status_code == 200:
            try:
                return resp.json()
            except requests.exceptions.JSONDecodeError:
                return {'text': resp.text}

        raise FailedAuth({
            'url': url, 'body': body, 'params': params, 'request': vars(resp.request), 'response': vars(resp)
        })

    @property
    def public_key(self) -> RSAPublicKey:
        if not self._public_key:
            public_b64_str = self._request('activeKey').get('publicKey')
            public_pem = self.b64_decode(public_b64_str)
            self._public_key = cereal.load_pem_public_key(public_pem)

        return self._public_key

    def validate(self, bat: str, raw=False) -> dict | None:
        """Method for validating and extracting data from an encoded BAT received from the server"""
        if not bat:
            return None

        comps = bat.split('.')
        pack = bat[0:bat.rfind('.')]
        signature = comps[2]

        try:
            sig = self.b64_decode(signature)
            self.public_key.verify(sig, pack.encode(), self.padding, self.sha256)
            full_data = {
                'headers': json.loads(self.b64_decode(comps[0])),
                'data': json.loads(self.b64_decode(comps[1]))
            }
            return full_data if raw else full_data['data']
        except InvalidSignature:
            return None

    def invalidate_ref(self, ref: str) -> None:
        """The backend request involved in logging out a user."""
        body = {**self.auth, 'data': {'ref': ref}}
        self._request('logout', body=body)

    def fetch_ref(self) -> str:
        """Method for obtaining a raw reference string from the authentication server for given client"""
        return self._request('credential', params=self.auth).get('ref')

    def build_auth(self) -> PreppedAuth:
        """Fetches a reference string and builds an associated login url"""
        ref = self.fetch_ref()
        return {'ref': ref, 'url': f"{self.server}?ref={ref}&callback={self.redirect_uri}"}

    def retrieve(self, ref: str):
        p = {'client_id': self.id, 'api_key': self.key, 'ref': ref}
        resp = self._request('retrieveToken', params=p, raw=True)
        if resp.status_code == 200:
            return resp.json()['bat']
        elif resp.status_code == 403:
            raise RevokedAuth({'ref': ref})

    def full_fetch(self, ref: str):
        return self.validate(self.retrieve(ref))



from .configure import config
from .cache import CachedRequester
from .redis_ import RedisMapping
from .pickles import Pickler
from .toml import edit_toml, load_toml, save_toml

__all__ = [
    'config', "CachedRequester", "Pickler", "RedisMapping",
    'edit_toml', 'load_toml', 'save_toml'
]
__version__ = "0.1.1rc13.dev1"
__name__ = "birdwell.mapping"
__title__ = "birdwell.mapping"



import string

chars = set(string.ascii_lowercase + string.digits)


def _tri_sub(word):
    padded = f'  {word} '
    return set(padded[i:i+3] for i in range(len(padded)-2))


def tri_gen(word):
    tri = set()
    sym_swapped = ''.join(c if c in chars else ' ' for c in word.lower())
    for sub in sym_swapped.split():
        if not sub:
            continue

        tri = tri.union(_tri_sub(sub))

    return frozenset(tri)


def tri_compare(word1: str, word2: str):
    """generate trigrams for and return overall similarity between them. order irrelevant"""
    return tri_calc(tri_gen(word1), tri_gen(word2))


def tri_compare_word_similarity(word1: str, word2: str):
    """generate trigrams for and return substring similarity between them. result is relative to word1"""
    return tri_calc_ws(tri_gen(word1), tri_gen(word2))


def tri_calc(tri1: set | frozenset, tri2: set | frozenset):
    """calculates overall similarity of two trigrams"""
    try:
        return len(tri1.intersection(tri2)) / len(tri1.union(tri2))
    except ZeroDivisionError:
        return 0.0


def tri_calc_ws(tri1: set | frozenset, tri2: set | frozenset):
    """calculates word similarity (substring analysis) of two trigrams"""
    if not (d := len(tri1)):
        return 0.0

    return len(tri1.intersection(tri2)) / d


def tri_calc_bws(tri1: set | frozenset, tri2: set | frozenset, threshold: float = .8):
    """bi-directional substring analysis"""
    return tri_calc_ws(tri1, tri2) >= threshold or tri_calc_ws(tri2, tri1) >= threshold




import pickle
import os
from contextlib import contextmanager


class Pickler:
    def __init__(self, root_dir: str = '.', file_ext: str = '.pickle'):
        self.loc = self.ensure_root(root_dir)
        self.file_ext = file_ext if file_ext.startswith('.') else f'.{file_ext}'
        self.data = self.load_all()

    @staticmethod
    def ensure_root(root_path: str):
        full_path = os.path.abspath(f'{root_path}/.birdwell/pickles')
        if not os.path.exists(full_path):
            os.makedirs(full_path, exist_ok=True)
        return full_path

    def save_all(self):
        for k in self.data:
            self.save(self.data[k], k)

    def load_all(self):
        loaded = {}
        for x in os.listdir(self.loc):
            if not x.endswith(self.file_ext):
                continue
            name = x.removesuffix(self.file_ext)
            loaded[name] = self.load(name)
        return loaded

    def load(self, name: str):
        try:
            with open(f'{self.loc}/{name}{self.file_ext}', mode='rb') as f:
                return pickle.load(f)
        except FileNotFoundError as e:
            return e
        except AttributeError as e:
            print(f'[PICKLER]: Error loading "{name}" - class attribute change detected. -> {str(e.args)}')
            return e

    def save(self, obj, name: str):
        if not isinstance(name, (str, int)):
            raise ValueError('name should be a string or int!')

        with open(f'{self.loc}/{name}{self.file_ext}', mode='wb') as f:
            pickle.dump(obj, f)

        self.data[name] = obj

    def rm(self, name: str):
        file_name = f'{name}{self.file_ext}'
        if file_name in os.listdir(self.loc):
            os.remove(f'{self.loc}/{file_name}')
        if name in self.data:
            self.data.pop(name)

    def clear(self):
        ks = [k for k in self.data]
        for k in ks:
            self.rm(k)


@contextmanager
def edit_pickle(file_path: str):
    with open(file_path, mode='rb') as f:
        loaded = pickle.load(f)

    try:
        yield loaded
    finally:
        with open(file_path, 'wb') as f2:
            pickle.dump(loaded, f2)



from __future__ import annotations
import threading
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Callable, TypeVar, Final, Generator, Concatenate, Self, Iterator, cast, ClassVar

from .items import fracture


def wait_threads(threads: list[threading.Thread], verbose=True):
    while len(threads):
        for idx in range(len(threads)-1, -1, -1):
            if not threads[idx].is_alive():
                threads.pop(idx)
                if verbose:
                    print(f'REMAINING THREADS {len(threads)}')


def thread_wrapped(func, items: list, thread_count=8, per_call=1,
                   output_mode=None, verbose=False, respawns=-1):
    """

    Parameters
    ----------
    func : function called with item chunks passed as only arg
    items : full list of items to process with func
    thread_count : number of threads to perform work with
    per_call : number of items to pass to func per call
    output_mode: func output handling
    verbose: print cycle / remaining / current on func call
    respawns: define max exception count; int used as a direct limit,
        float value used to calculate limit using the formula: respawns = int(len(items) // per_call * respawns)

    """

    if not isinstance(per_call, int) or per_call <= 0:
        raise ValueError('Items per call must be positive int')

    running = True  # used to stop threads via main
    threads = []
    packs = items[::] if per_call == 1 else list(fracture(items, per_call))
    cycles, revives = 0, 0

    # calculate scaling respawns
    if isinstance(respawns, float):
        respawns = int(len(items) // per_call * respawns)

    # func is wrapped to give calling thread access to context
    def wrapped_func():
        # thread will continue running until work is complete or interrupted
        while packs and running:
            item = packs.pop()
            nonlocal cycles
            cycles += 1
            if verbose:
                print(f'CYCLE: {cycles:_}, REMAIN: {len(packs):_} CURRENT: {item}')

            res = func(item)
            if output_mode == 'append' and res:
                packs.extend(res)

    # initial thread population
    for i in range(thread_count):
        tr = threading.Thread(target=wrapped_func, daemon=True)
        threads.append(tr)
        tr.start()

    # try/except used for shutdown control
    try:
        # monitor thread progress and block main
        # note: could also make this a threaded operation to not block
        while threads:
            # reversed index used to enable popping
            for idx in range(len(threads)-1, -1, -1):
                # remove as completed/errored
                if not threads[idx].is_alive():
                    threads.pop(idx)

            # ensure threadpool runs at full capacity while items are available
            if respawns is not None and running and (len(threads) < thread_count < len(packs)):
                if respawns == -1 or respawns - revives >= 0:

                    etc = '' if not verbose else (
                        f'REVIVE_COUNT=[{revives + 1}] '
                        f'REMAINING_RESPAWNS=[{'UNLIMITED' if respawns == -1 else respawns - revives}] '
                        f'SPAWNING REPLACEMENT.')

                    print(f'EARLY THREAD TERMINATION DETECTED. {etc}')
                    # replace dead threads
                    tr = threading.Thread(target=wrapped_func, daemon=True)
                    threads.append(tr)
                    tr.start()
                    revives += 1

    # propagate stop/interrupt to inner threads
    except KeyboardInterrupt as e:
        running = False
        raise e


@dataclass
class BackgroundCall[_R]:
    result: _R | None
    thread: threading.Thread | None


class BackgroundThread[**_P, _R]:
    def __init__(self, func: Callable[_P, _R], raise_exceptions=False, timeout=None):
        self.func: Callable[_P, _R] = func
        self.data: BackgroundCall[_R] = BackgroundCall(result=None, thread=None)
        # self.result: _R = None
        # self.thread: threading.Thread | None = None
        self.raise_exceptions = raise_exceptions
        self.timeout = timeout
    #
    # def wrapped(self, *args: _P.args, **kwargs: _P.kwargs) -> None:
    #     try:
    #         self.data.result = self.func(*args, **kwargs)
    #     except Exception as e:
    #         self.data.result = e

    @contextmanager
    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> Generator[BackgroundCall[_R], None, None]:
        def inner_call():
            try:
                self.data.result = self.func(*args, **kwargs)
            except Exception as e:
                self.data.result = e

        self.data.thread = threading.Thread(target=inner_call, daemon=True)
        self.data.thread.start()
        try:
            yield self.data
        finally:
            self.data.thread.join(timeout=self.timeout)
            if self.raise_exceptions and isinstance(self.data.result, Exception):
                raise self.data.result


@contextmanager
def background_thread[**_P, _R](func: Callable[[_P], _R], *args: _P.args, **kwargs: _P.kwargs):
    result: _R = None
    res: BackgroundCall = BackgroundCall(result=result, thread=None)

    def wrapped():
        try:
            res.result = func(*args, **kwargs)
        except Exception as e:
            res.result = e

    res.thread = threading.Thread(target=wrapped, daemon=True)
    res.thread.start()
    try:
        yield res
    finally:
        res.thread.join()



from configparser import ConfigParser
from os import environ as env


def config(file_path: str, section: str, env_overrides=False):
    """
    Parser for pulling config dict from a file, with optional env overriding to modify base configs or take their place.
    Environmental variables should take the form 'CONFIG_{section}_{key} = {value}'

    Parameters
    ----------
    file_path : str
        config file location
    section : str
        header name within config file
    env_overrides : bool
        allow environmental variables to update or extend base config pulled from provided file

    Returns
    -------
    dict
        containing key value pairs matching specified section of config file
    """

    parser = ConfigParser()
    parser.read(file_path)
    configs = {}

    if parser.has_section(section):
        # item tuples returned as dict
        configs = {k: v for k, v in parser.items(section)}
    if env_overrides:
        prefix = f'CONFIG_{section.upper}_'
        overrides = {k.removeprefix(prefix).lower(): v for k, v in env.items() if k.startswith(prefix)}
        configs = configs | overrides

    if configs:
        return configs

    raise Exception(f'Config missing for {section}')
